#pragma once

#include <ThreadPool/Mutex.h>
#include <tommath.h>
#include <memory>
#include <deque>

namespace ts {
    namespace server {
        class ConnectedClient;
    }

    namespace protocol {
        struct Puzzle {
            mp_int x;
            mp_int n;
            int level;

            mp_int result;

            uint8_t data_x[64];
            uint8_t data_n[64];
            uint8_t data_result[64];
        };
        class PuzzleManager {
            public:
                PuzzleManager();
                ~PuzzleManager();

                bool precomputePuzzles(size_t limit);

                size_t precomputedPuzzleCount();

                std::shared_ptr<Puzzle> nextPuzzle();
            private:
                void generatePuzzle();

                threads::Mutex indexLock;
                size_t cacheIndex = 0;

                std::deque<std::shared_ptr<Puzzle>> cached;
        };
    }
}