#pragma once

#ifdef COMPILE_WEB_CLIENT
#include <pipes/ssl.h>
#include <pipes/ws.h>
#include <pipes/rtc/PeerConnection.h>
#include <src/client/SpeakingClient.h>
#include <src/client/ConnectedClient.h>
#include <protocol/buffers.h>
#include "misc/queue.h"
#include "SampleHandler.h"
#include <opus/opus.h>
#include "VoiceBridge.h"
#include <json/json.h>
#include <EventLoop.h>

namespace ts {
    namespace server {
        class WebControlServer;

        class WebClient : public SpeakingClient {
                friend class WebControlServer;
            public:
                WebClient(WebControlServer*, int socketFd);
                ~WebClient() override;

                void sendJson(const Json::Value&);
                void sendCommand(const ts::Command &command, bool low = false) override;
                void sendCommand(const ts::command_builder &command, bool low) override;

                bool disconnect(const std::string &reason) override;
                bool close_connection(const std::chrono::system_clock::time_point& timeout = std::chrono::system_clock::time_point()) override;

                bool shouldReceiveVoice(const std::shared_ptr<ConnectedClient> &sender) override;

                inline std::chrono::nanoseconds client_ping() { return this->client_ping_layer_7(); }
                inline std::chrono::nanoseconds client_ping_layer_5() { return this->ping.value; }
                inline std::chrono::nanoseconds client_ping_layer_7() { return this->js_ping.value; }
            protected:
                void handlePacketVoiceWhisper(const pipes::buffer_view &string, bool) override;

            protected:
                void tick(const std::chrono::system_clock::time_point&) override; /* Every 500ms */

                void applySelfLock(const std::shared_ptr<WebClient> &cl){ _this = cl; }
            private:
                WebControlServer* handle;
                std::chrono::time_point<std::chrono::system_clock> connectedTimestamp;

                std::shared_mutex voice_bridge_lock;
                std::unique_ptr<web::VoiceBridge> voice_bridge;
                int file_descriptor;

                std::mutex event_lock;
                ::event* readEvent;
                ::event* writeEvent;

                struct {
                    uint8_t current_id = 0;
                    std::chrono::system_clock::time_point last_request;
                    std::chrono::system_clock::time_point last_response;

                    std::chrono::nanoseconds value{};
                    std::chrono::nanoseconds timeout{2000};
                } ping;

                struct {
                    uint8_t current_id = 0;
                    std::chrono::system_clock::time_point last_request;
                    std::chrono::system_clock::time_point last_response;

                    std::chrono::nanoseconds value{};
                    std::chrono::nanoseconds timeout{2000};
                } js_ping;

                std::mutex queue_lock;
                std::deque<pipes::buffer> queue_read;
                std::deque<pipes::buffer> queue_write;
                threads::Mutex execute_lock; /* needs to be recursive! */

                std::thread flush_thread;
                std::recursive_mutex close_lock;
            private:
                void initialize();

                bool ssl_detected = false;
                bool ssl_encrypted = true;
                pipes::SSL ssl_handler;
                pipes::WebSocket ws_handler;
                void handleMessageRead(int, short, void*);
                void handleMessageWrite(int, short, void*);
                void enqueue_raw_packet(const pipes::buffer_view& /* buffer */);

                void processNextMessage(const std::chrono::system_clock::time_point& /* scheduled */);
                void registerMessageProcess();

                std::shared_ptr<event::ProxiedEventEntry<WebClient>> event_handle_packet;

                //WS events
                void onWSConnected();
                void onWSDisconnected(const std::string& reason);
                void onWSMessage(const pipes::WSMessage&);
            protected:
                void disconnectFinal();
                void handleMessage(const std::string &);

            public:
                void send_voice_packet(const pipes::buffer_view &view, const VoicePacketFlags &flags) override;
                void send_voice_whisper_packet(const pipes::buffer_view &view, const VoicePacketFlags &flags) override;

            protected:

                command_result handleCommand(Command &command) override;
                command_result handleCommandClientInit(Command &command) override;
        };
    }
}
#endif