#include "log/LogUtils.h"
#include "WebIoManager.h"
#include "src/Configuration.h"

using namespace std;
using namespace ts::webio;

void EventLoop::invoker_loop() {
    const timeval tick_timeout{10, 0};
    const auto original_base = this->loop;
    while(this->loop == original_base) {
        event_base_loopexit(original_base, &tick_timeout);
        event_base_dispatch(original_base);
    }
    debugMessage(LOG_GENERAL, "Event base dispatch of {} ended.", (void*) original_base);
}

LoopManager::LoopManager() {
    lock_guard lock(this->loop_lock);
    for(int i = 0; i < config::threads::web::io_loops; i++) {
        auto loop = make_shared<EventLoop>();
        loop->loop = event_base_new();
        loop->invoker = make_unique<thread>(&EventLoop::invoker_loop, loop.get());

        pthread_setname_np(loop->invoker->native_handle(), ("Web IO #" + to_string(i)).c_str());

        this->loops.push_back(move(loop));
    }
}

LoopManager::~LoopManager() {
    lock_guard lock(this->loop_lock);
    for(const auto& loop : this->loops) {
        const auto base = loop->loop;
        loop->loop = nullptr;

        const timeval timeout_now{0, 1};
        event_base_loopexit(base, &timeout_now);

        if(loop->invoker && loop->invoker->joinable())
            loop->invoker->join();

        event_base_free(base);
    }
}

std::shared_ptr<EventLoop> LoopManager::next_loop() {
    lock_guard lock(this->loop_lock);
    if(this->loops.empty())
        return nullptr;

    return this->loops[this->loop_index++ % this->loops.size()];
}